#pragma once

#define LIBOSMOCORE_VERSION {{VERSION}}
#define LIBOSMOCORE_VERSION_STR "{{VERSION}}"

#define LIBOSMOCORE_VERSION_MAJOR {{VERSION_MAJOR}}
#define LIBOSMOCORE_VERSION_MINOR {{VERSION_MINOR}}
#define LIBOSMOCORE_VERSION_PATCH {{VERSION_PATCH}}

#define LIBOSMOCORE_VERSION_GREATER_EQUAL(major, minor, patch) \
	(LIBOSMOCORE_VERSION_MAJOR > (major) || \
	 (LIBOSMOCORE_VERSION_MAJOR == (major) && \
	  LIBOSMOCORE_VERSION_MINOR > (minor)) || \
	 (LIBOSMOCORE_VERSION_MAJOR == (major) && \
	  LIBOSMOCORE_VERSION_MINOR == (minor) && \
	  LIBOSMOCORE_VERSION_PATCH >= (patch)))
